
/*
he2mu.c

This program read metadata from HDF-EOS file and write metadata 
back to the HDF-EOS file.

Zhangshi Yin
 
2003.3.31
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "mfhdf.h"
#include "odldef.h"

/* Read metadata from HDF. */
int ReadMetadata(int32 fid, char *MetadataName);

/* Write metadata to HDF. */ 
int WriteMetadata(int32 fid, FILE *MetadataFile, FILE *HDFMetadataFile);

/* Write metadata group. */
int WriteGroup(int32 fid, FILE *MFile, char *MetadataName, char *GroupName);

/* Iterate all metadata. */
intn IterateMetadata(int32 fid);

/* Reset metadata. */
intn ResetMetadata(int32 fid);

intn ResetAttribute(int32 fid, char *MetadataName);

/* Get a value of ODL attribute. */
int GetValue(FILE *metadataFile, char *GroupName, char *attributeName, char *valueName, char *returnValue);

/* Check INVENTORYMETADATA, COREMETADATA, COLLECTIONMETADATA groups */
int CheckMetaGroup(FILE *metadataFile,  char *GroupName);

/* Check values are equal. */
void CheckValue(FILE *HDFMetadataFile, FILE *MetadataFile, char *GroupName);

/* Print help message. */
/*
void PrintHelp(void);
*/

/* Convert low case to upcase. */ 
void LowToUpper (char *inLine, char *outLine);

/* Convert space line to no space line. */
void NoSpace(char *inLine, char *outLine);

FILE *TempFile;
int Group[3];
char ReadReplaceMerge='n';

int he2mu(int argc, char *argv[])
{
  int32 fid;
  int  i, j;
  int FlagLength, DataFile=0, MetaFile=0;
  char HDFEOSFileName[256];
  char MetadataFileName[256];
  AGGREGATE root;
  int status = 0;
  char aLine[1024];
  int HDFGroup, METGroup;
  FILE *MetadataFile, *HDFMetadataFile;

  /*
  Get comand-line argument.
  */
  if(argc==1)
  {
    PrintHelp();
  }
  for (i=1; i<argc; i++)
  {
    if (argv[i][0]=='-')
    {
      FlagLength=strlen(argv[i]);
      for (j=1; j<FlagLength; j++)
      {
        switch(argv[i][j])
        {
          case 'h':
              PrintHelp();
              exit(1);
          case 'x':
              ReadReplaceMerge='x';
	      if(i<(argc-1)) 
              {
		strcpy(MetadataFileName, argv[i+1]);
		fprintf(stderr,"Metadata file: %s.\n", MetadataFileName);
		MetaFile=1;
		i++;
	      }
              break;
          case 'f':
              ReadReplaceMerge='f';
	      if(i<(argc-1)) 
              {
		strcpy(MetadataFileName, argv[i+1]);
		fprintf(stderr,"Metadata file: %s.\n", MetadataFileName);
		MetaFile=1;
		i++;
	      }
              break;
          case 'm':
              ReadReplaceMerge='m';
	      if(i<(argc-1)) 
              {
		strcpy(MetadataFileName, argv[i+1]);
		fprintf(stderr,"Metadata file: %s.\n", MetadataFileName);
		MetaFile=1;
		i++;
	      }
              break;
	default:
	  fprintf(stderr,"Unknown option %s\n", argv[i]);
	  exit(1);
        }
      }
    }
    else 
    {
      DataFile = 1;
      strcpy(HDFEOSFileName, argv[i]);
      fprintf(stderr, "HDFEOS file: %s.\n", HDFEOSFileName);
    }
  }

  if (DataFile==0)
  {
    fprintf(stderr, "No HDFEOS file.\n");
    PrintHelp();
    exit(1);
  }

  if (ReadReplaceMerge=='n')
  {
    fprintf(stderr, "Must specify one of -x|-f|-m.\n");
    PrintHelp();
    exit(1);
  }

  /* 
  Read metadata from HDF file.  
  */
  if (ReadReplaceMerge=='x')
  {
    /* Open HDFEOS file. */
    fid = SDstart(HDFEOSFileName, DFACC_READ);
    if (fid == -1)
    {
      fprintf(stderr, "Unable to open HDF-EOS file %s \n", HDFEOSFileName);
      exit(1);
    }

    /* Open metadata file. */
    if(MetaFile == 1)
    {
      MetadataFile=fopen(MetadataFileName, "w+");
    }
    else
    {
      MetadataFile = stdout;
    }

    if (MetadataFile==NULL)
    {
      fprintf(stderr, "Cannot open metadata file %s.\n", MetadataFileName);
      exit(1);
    }

    fprintf(stderr,"It is dumping metadata ...\n"); 

    /*
    Iterate through the file.
    */
    TempFile=tmpfile();
    if (IterateMetadata(fid)!=0)
    {
      fprintf(stderr, "Cannot iterate HDFEOS file.\n");
      exit(1);
    }

    fseek(TempFile, 0, 0);
    while (fgets(aLine, 1024, TempFile))
    {
      if (strcmp(aLine, "END\n")!=0 && strcmp(aLine, "END")!=0)
      {
        fprintf(MetadataFile, "%s", aLine); 
      }
    }
    fprintf(MetadataFile, "END\n\n"); 
    fclose(TempFile);

    /*
    Check syntax of MetadataFile.
    */
    /*
    fseek(MetadataFile, 0, 0);
    root=NewAggregate(NULL, KA_GROUP, "test", NULL);
    status=ReadLabel(MetadataFile, root);
    if (status!=1)
    {
      fprintf(stderr, "There are syntax errors in HDF metadata file.\n");
      exit(1);
    }
    RemoveAggregate(root);
    */

    if (fclose(MetadataFile)==-1)
    {
      fprintf(stderr,"Cannot close metadata file %s.\n", MetadataFileName);
      exit(1);
    }

    if (SDend(fid)==-1)
    {
      fprintf(stderr,"Cannot close HDF-EOS file.\n");
      exit(1);
    }

    fprintf(stderr,"Finished metadata dumping.\n");
  }

  /*
  Write metadata to HDF file.
  */
  if (ReadReplaceMerge=='f' || ReadReplaceMerge=='m')
  {
    /* Open HDFEOS file. */
    fid = SDstart(HDFEOSFileName, DFACC_RDWR);
    if (fid == -1)
    {
      fprintf(stderr, "Unable to open HDF-EOS file %s \n", HDFEOSFileName);
      exit(1);
    }

    /* Open metadata file. */
    if(MetaFile == 1)
    {
      MetadataFile=fopen(MetadataFileName, "r");
    }
    else 
    {
      /* 
      metadata file is stdin, so read it into a temporary file.
      */
      MetadataFile = tmpfile();
      while ((i= fread(aLine, 1, 1024, stdin)) > 0)
      {
	fwrite(aLine, 1, 1024, MetadataFile);
      }
      fseek(MetadataFile, 0, 0);
    }

    if (MetadataFile==NULL)
    {
      fprintf(stderr, "Cannot open metadata file %s.\n", MetadataFileName);
      exit(1);
    }

    /*
    Check syntax of MetadataFile.
    */
    fseek(MetadataFile, 0, 0);
    root=NewAggregate(NULL, KA_GROUP, "test", NULL);
    status=ReadLabel(MetadataFile, root);
    if (status!=1)
    {
      fprintf(stderr, "There are syntax errors in input metadata file.\n");
      exit(1);
    }
    RemoveAggregate(root);

    /*
    Read HDF metadata.
    */
    TempFile=tmpfile();
    if (IterateMetadata(fid)!=0)
    {
      fprintf(stderr, "Cannot iterate HDFEOS file.\n");
      exit(1);
    }

    /* Process TempFile to HDFMetadataFile. */
    fseek(TempFile, 0, 0);
    HDFMetadataFile=tmpfile();
    while (fgets(aLine, 1024, TempFile))
    {
      if (strcmp(aLine, "END\n")!=0 && strcmp(aLine, "END")!=0)
      {
        fprintf(HDFMetadataFile, "%s", aLine); 
      }
    }
    fprintf(HDFMetadataFile, "END\n\n"); 
    
    fclose(TempFile);
   
    /*
    Check syntax of MetadataFile.
    */
    fseek(HDFMetadataFile, 0, 0);
    root=NewAggregate(NULL, KA_GROUP, "test", NULL);
    status=ReadLabel(HDFMetadataFile, root);
    if (status!=1)
    {
      fprintf(stderr, "There are syntax errors in HDF-EOS metadata file.\n");
      exit(1);
    }
    RemoveAggregate(root);

    /* Check COLLECTION, INVENTORY, ARCHIVE groups. */
    Group[0]=0;
    Group[1]=0;
    Group[2]=0;
    if (ReadReplaceMerge=='m')
    {
      Group[0]=CheckMetaGroup(MetadataFile, "COLLECTIONMETADATA");
      Group[1]=CheckMetaGroup(MetadataFile, "INVENTORYMETADATA");
      Group[2]=CheckMetaGroup(HDFMetadataFile, "ARCHIVEMETADATA");
    }

    /* Check INVENTORY values. */
    METGroup=CheckMetaGroup(MetadataFile, "INVENTORYMETADATA");
    HDFGroup=CheckMetaGroup(HDFMetadataFile, "INVENTORYMETADATA");
    if ((HDFGroup==1) && (METGroup==1))
    {
      CheckValue(HDFMetadataFile, MetadataFile, "INVENTORYMETADATA");
    }

    /*
    Reset HDF metadata to NULL.
    */
    if (ResetMetadata(fid)!=0)
    {
      fprintf(stderr, "Cannot reset metadata.\n");
      exit(1);
    }

    WriteMetadata(fid, MetadataFile, HDFMetadataFile);

    if (fclose(MetadataFile)==-1)
    {
      fprintf(stderr,"Cannot close metadata file.\n");
      exit(1);
    }
    if (fclose(HDFMetadataFile)==-1)
    {
      fprintf(stderr,"Cannot close HDFMetadata file.\n");
      exit(1);
    }

    if (SDend(fid)==-1)
    {
      fprintf(stderr,"Cannot close HDF-EOS file.\n");
      exit(1);
    }

    fprintf(stderr,"Finished metadata writing.\n"); 
  }
  return 0;
}

/* Read metadata from HDF-EOS file and write to temp file. */
intn IterateMetadata(int32 fid)
{
  intn i, j;
  char MetadataName[100];
  char MetadataNames[100][100]={"CoreMetadata", "COREMETADATA", "coremetadata",
                                "ArchiveMetadata", "ARCHIVEMETADATA", "archivemetadata",
                                "CollectionMetadata", "COLLECTIONMETADATA", "collectionmetadata"
                               };

  for (i = 0; i <= 9; i++)
  {
    ReadMetadata(fid, MetadataNames[i]);
    for (j = 0; j <= 9; j++)
    {
      sprintf(MetadataName, "%s.%d", MetadataNames[i], j);
      ReadMetadata(fid, MetadataName);
    }
  }

  return 0;
}

/* Reset metadata to NULL. */
intn ResetMetadata(int32 fid)
{
  intn i, j;
  char MetadataName[100];
  char MetadataNames[100][100]={"CoreMetadata", "COREMETADATA", "coremetadata",
                                "ArchiveMetadata", "ARCHIVEMETADATA", "archivemetadata",
                                "CollectionMetadata", "COLLECTIONMETADATA", "collectionmetadata"
                               };

  for (i = 0; i <= 9; i++)
  {
    ResetAttribute(fid, MetadataNames[i]);
    for (j = 0; j <= 9; j++)
    {
      sprintf(MetadataName, "%s.%d", MetadataNames[i], j);
      ResetAttribute(fid, MetadataName);
    }
  }

  return 0;
}

/* Read metadata from HDFEOS and write temp file. */
intn ReadMetadata(int32 fid, char *MetadataName)
{
  intn status = 0;
  int32 AttributeIndex, DataType, TotalSize;
  char TempString[50], *AString, *Metadata, AttributeName[MAX_NC_NAME];

  /* Look for attribute */
  AttributeIndex = SDfindattr(fid, MetadataName);
  if (AttributeIndex == -1)
  {
    return (1);
  }
  /* Get size of attribute  */
  status = SDattrinfo(fid, AttributeIndex, AttributeName, &DataType, &TotalSize);
  /* size too small? */
  /*TotalSize+=1000;*/
  if (status == -1)
  {
    return (1);
  }

  /* Allocate memory for attribute contents */
  Metadata = (char *)calloc(TotalSize, 1);
  if (Metadata == NULL)
  {
    fprintf(stderr, "Unable to allocate %d bytes for %s.\n", TotalSize, MetadataName);
    exit(1);
  }

  /* Read attribute */
  status = SDreadattr(fid, AttributeIndex, Metadata);
  if (status == -1)
  {
    free(Metadata);
    return (1);
  }

  if (ReadReplaceMerge=='x')
  {
    fprintf(stderr, "Extracting metadata %s.\n", MetadataName);
  }

  fprintf(TempFile, "%s", Metadata);

  free(Metadata);

  return (0);
}

/* Reset attribute. */
intn ResetAttribute(int32 fid, char *MetadataName)
{
  intn status = 0;
  int32 AttributeIndex, DataType, TotalSize, count, DataTypeSize;
  char  AttributeName[MAX_NC_NAME];
  void *Metadata;

  /* Look for attribute */
  AttributeIndex = SDfindattr(fid, MetadataName);
  if (AttributeIndex == -1)
  {
    return (1);
  }
  /* Get size of attribute  */
  status = SDattrinfo(fid, AttributeIndex, AttributeName, &DataType, &TotalSize);
  if (status == -1)
  {
    return (1);
  }

  /* Allocate memory for attribute contents */
  Metadata = (void *)calloc(TotalSize, 1);
  if (Metadata == NULL)
  {
    fprintf(stderr, "Unable to allocate %d bytes for %s.\n", TotalSize, MetadataName);
    exit(1);
  }

switch(DataType)
{
  case DFNT_INT8:
  case DFNT_UINT8:
  case DFNT_UCHAR8:
  case DFNT_CHAR8:
  case DFNT_NINT8:
  case DFNT_NUINT8:
  case DFNT_NCHAR8:
  case DFNT_NUCHAR8:
    DataTypeSize=1;
    break;
  case DFNT_NCHAR16:
  case DFNT_NUCHAR16:
  case DFNT_CHAR16:
  case DFNT_UCHAR16:
  case DFNT_NINT16:
  case DFNT_NUINT16:
  case DFNT_INT16:
  case DFNT_UINT16:
    DataTypeSize=2;
    break;
  case DFNT_FLOAT32:
  case DFNT_INT32:
  case DFNT_UINT32:
  case DFNT_NINT32:
  case DFNT_NUINT32:
  case DFNT_NFLOAT32:
    DataTypeSize=4;
    break;
  case DFNT_FLOAT64:
  case DFNT_INT64:
  case DFNT_UINT64:
  case DFNT_NINT64:
  case DFNT_NUINT64:
  case DFNT_NFLOAT64:
    DataTypeSize=8;
    break;
}
  count=TotalSize/DataTypeSize;

  /* Read attribute */
  status = SDsetattr(fid, MetadataName, DataType, count, Metadata);
  if (status == -1)
  {
    free(Metadata);
    return (1);
  }

  free(Metadata);

  return (0);
}

/* Write HDF metadata. */ 
int WriteMetadata(int32 fid, FILE *MetadataFile, FILE *HDFMetadataFile)
{ 
  fprintf(stderr,"It is writing metadata ...\n");

  /* 
  Read coremetadata, archivemetadata and collectionmetadata 
  by find the groups. If more than 65535, separate it.
  */

  /* Replace. */
  if (ReadReplaceMerge=='f')
  {
    WriteGroup(fid, MetadataFile, "CollectionMetadata", "COLLECTIONMETADATA");
    WriteGroup(fid, MetadataFile, "CoreMetadata", "INVENTORYMETADATA");
    WriteGroup(fid, MetadataFile, "ArchiveMetadata", "ARCHIVEMETADATA");
  }
  /* Merge. */
  if (ReadReplaceMerge=='m')
  {
    WriteGroup(fid, MetadataFile, "CollectionMetadata", "COLLECTIONMETADATA");
    WriteGroup(fid, MetadataFile, "CoreMetadata", "INVENTORYMETADATA");
    WriteGroup(fid, HDFMetadataFile, "ArchiveMetadata", "ARCHIVEMETADATA");
  }

  return 0;
}

/* Write MetadataFile to hdfFile. */
int WriteGroup(int32 fid, FILE *MFile, char *MetadataName, char *GroupName)
{
  intn i, strsize;
  int32 TotalSize;
  char *MetaBuffer, *MetaPoint;
  char aLine[1024];
  char MetadataNamen[256];
  int  status;
  AGGREGATE root, aAggregate;
  FILE *GroupFile;

  fseek(MFile, 0, 0);
  root=NewAggregate(NULL, KA_GROUP, "test", NULL);
  ReadLabel(MFile, root);
  aAggregate=FindGroup(root, GroupName);
  GroupFile=tmpfile();
  WriteODL(GroupFile, aAggregate);

  /*
  From the User's Guide
  Maximum length of a vdata record: 65535

  From ./hdf/src/hlimits.h
  Maximum order size:  65535
  MAX_FIELD_SIZE: 65535
  Max name length: 65
  Max NO. of fields: 256
  MAX Field Name length: 128
  */

  MetaBuffer=(char *)calloc(65535, 1);
  MetaPoint=MetaBuffer;

  i=-1;
  TotalSize=0;
  fseek(GroupFile, 0, 0);
  while (fgets(aLine, 1024, GroupFile))
  {
    strsize=strlen(aLine);
    TotalSize+=strsize;
    if (TotalSize>65531) /* Need to add "END\n", so 65535-4=65531 */
    {
      TotalSize-=strsize;

  if (i==-1)
  {
    sprintf(MetadataNamen, "%s", MetadataName);
  }
  else
  {
    sprintf(MetadataNamen, "%s.%d", MetadataName, i);
  }

      SDsetattr(fid, MetadataNamen, 4, TotalSize, MetaBuffer);
      fprintf(stderr, "Write metadata %s.\n", MetadataNamen);

      free(MetaBuffer);

      i++;
      TotalSize=strsize;
      MetaBuffer=(char *)calloc(65535, 1);
      MetaPoint=MetaBuffer;
    }
    memcpy(MetaPoint, aLine, strsize);
    MetaPoint+=strsize;
  }
  memcpy(MetaPoint, "END\n", 4);
  MetaPoint+=4;
  TotalSize+=4;

  if (i==-1)
  {
    sprintf(MetadataNamen, "%s", MetadataName);
  }
  else
  {
    sprintf(MetadataNamen, "%s.%d", MetadataName, i);
  }

  SDsetattr(fid, MetadataNamen, 4, TotalSize, MetaBuffer);
  fprintf(stderr, "Write metadata %s.\n", MetadataNamen);

  free(MetaBuffer);

  fclose(GroupFile);

  return (0);
}

/* Get value of ODL attribute. */
int GetValue(FILE *metadataFile, char *GroupName,  char *attributeName, char *valueName, char *returnValue)
{
  AGGREGATE root, aAggregate, bAggregate;
  PARAMETER aParameter;
  VALUE aValue;
  int  status;

  fseek(metadataFile, 0, 0);
  root=NewAggregate(NULL, KA_GROUP, "test", NULL);

  /* Read odl, no error return 1 and error return other. */
  /* fprintf(stderr,"Reading and checking odl.\n"); */
  status=ReadLabel(metadataFile, root);
  if (status!=1)
  {
    fprintf(stderr, "There are syntax errors in metadata file.\n");
    exit(1);
  }
  bAggregate=FindGroup(root, GroupName);
  if (bAggregate==NULL)
  {
    sprintf(returnValue, "No group");
    return (0);
  }
  aAggregate=FindAggregate(bAggregate, attributeName);
  if (aAggregate==NULL)
  {
    sprintf(returnValue, "No attribute");
    return (0);
  }

  /* Find parameter. */
  aParameter=FindParameter(aAggregate, valueName);
  if (aParameter==NULL)
  {
    sprintf(returnValue, "No parameter");
    return (0);
  }

  /* Check value. */
  aValue=FirstValue(aParameter);
  if (aValue==NULL)
  {
    sprintf(returnValue, "No value");
    return (0);
  }

  if (aValue->item.type==TV_STRING)
  {
    /* print string and time. */
    sprintf(returnValue, "%s", aValue->item.value.string);
  }
  if (aValue->item.type==TV_INTEGER)
  {
    /* print integer. */
    sprintf(returnValue, "%i", aValue->item.value.integer.number);
  }

  RemoveAggregate(root);

  return (0);
}

/* Print help message. */
/*
void PrintHelp(void)
{
  fprintf(stderr,"\nUsage: \n");
  fprintf(stderr,"he2mu HDFEOSFileName -xmf [MetadataFileName]\n\n");
  fprintf(stderr,"   HDFEOSFileName--HDFEOS file. Required.\n");
  fprintf(stderr,"   -x Extract metadata from HDFEOS file.\n");
  fprintf(stderr,"   -m Update metadata (merge current metadata with input). \n");
  fprintf(stderr,"   -f Forcibly replace metadata with metadata from input.\n");
  fprintf(stderr,"   MetadataFileName--Metadata file. Optional (default is standard input/output).\n");
  fprintf(stderr,"   -h Print help message.\n");
  fprintf(stderr,"\n");
  exit(1);
}
*/

/* Convert low case to upper case. */ 
void LowToUpper(char *inLine, char *outLine)
{
  unsigned int i;
  char astring, bstring;
  unsigned int ilen = strlen(inLine);

  for (i=0; i<ilen; i++)
  {
    astring=inLine[i];
    bstring=toupper(astring);
    outLine[i]=bstring;
  }
  outLine[ilen]='\0';
}

/* Change line with space to no space line. */
void NoSpace(char *inLine, char *outLine)
{
  char *astring, *bstring;

  /* Get no space line. */
  strcpy(outLine, "");
  astring=strtok(inLine, "\n");
  bstring=strtok(astring, "	");
  astring=strtok(bstring, " ");
  while(astring)
  {
    strcat(outLine, astring);
    astring=strtok(NULL, " ");
  }
}

/* Check INVENTORYMETADATA, COREMETADATA, COLLECTIONMETADATA groups */
int CheckMetaGroup(FILE *metadataFile,  char *GroupName)
{
  AGGREGATE root, aAggregate;
  int  status;

  fseek(metadataFile, 0, 0);
  root=NewAggregate(NULL, KA_GROUP, "test", NULL);

  status=ReadLabel(metadataFile, root);
  if (status!=1)
  {
    fprintf(stderr, "There are syntax errors in metadata file.\n");
    exit(1);
  }
  aAggregate=FindGroup(root, GroupName);
  if (aAggregate==NULL)
  {
     return 0;
  }
  else
  {
     return 1;
  }
}

void CheckValue(FILE *HDFMetadataFile, FILE *MetadataFile, char *GroupName)
{
  char HDFValue[256], METValue[256], attributeName[256], valueName[256];
  char ParameterName[3][20]={"SHORTNAME", "PRODUCTIONDATETIME", "VERSIONID"};
  int i;

  for (i=0; i<3; i++)
  {
    fseek(HDFMetadataFile, 0, 0);
    fseek(MetadataFile, 0, 0);
    strcpy(attributeName, ParameterName[i]);
    strcpy(valueName, "VALUE");
    GetValue(MetadataFile, GroupName, attributeName, valueName, METValue);
    GetValue(HDFMetadataFile, GroupName, attributeName, valueName, HDFValue);
    if (strcmp(HDFValue, METValue)!=0)
    {  
      fprintf(stderr,"%s values: %s and %s in metadata file and HDFEOS file don't match.\n",
	      attributeName, METValue, HDFValue);
      exit(1);
    }
  }
}

