/*
hemu.c

This program extracts HDF-EOS data objects such as grid, swath, point and their fields to 
binary and ascii files in HTML or XML index information. 

Zhangshi Yin

August 18, 2003.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "mfhdf.h"
#include "hdf5.h"
#include "odldef.h"

void PrintHelp();

main(int argc, char *argv[])
{
  char       *FileName;
  int32      hdfeos2fid;
  hid_t      hdfeos5fid; 

  /* 
  Check and get argument. 
  */
  if(argc==1)
  {
    PrintHelp();
  }
  FileName=argv[1];

  /*
  Check file format.
  */

  hdfeos2fid = SDstart(FileName, DFACC_READ);
  if (hdfeos2fid != FAIL)
  {
    if (SDend(hdfeos2fid) == FAIL)
    {
      printf("Cannot close file.\n");
      exit(1);
    }
    he2mu(argc, argv);
  }
  else
  {
    hdfeos5fid=H5Fopen(FileName, H5F_ACC_RDONLY, H5P_DEFAULT);
    if (hdfeos5fid >=0) 
    {
      if (H5Fclose(hdfeos5fid) == FAIL)
      {
        printf("Cannot close file.\n");
        exit(1);
       }
       he5mu(argc, argv);
    }
    else
    {
      printf("File cannot be opened.\n");
      PrintHelp();
      exit(1);
    }
  }

  return(0);
}

void PrintHelp()
{
  fprintf(stderr,"\nUsage: \n");
  fprintf(stderr,"hemu HDFEOSFileName -xmf [MetadataFileName]\n\n");
  fprintf(stderr,"   HDFEOSFileName--HDFEOS file. Required.\n");
  fprintf(stderr,"   -x Extract metadata from HDFEOS file.\n");
  fprintf(stderr,"   -m Update metadata (merge current metadata with input). \n");
  fprintf(stderr,"   -f Forcibly replace metadata with metadata from input.\n");
  fprintf(stderr,"   MetadataFileName--Metadata file. Optional (default is standard input/output).\n");
  fprintf(stderr,"   -h Print help message.\n");
  fprintf(stderr,"\n");
  exit(1);
}

